//+------------------------------------------------------------------+
//|                          Divergence for Many Indicators v4_0.mq5 |
//|                                        Copyright 2023, PuguForex |
//|                          https://www.mql5.com/en/users/puguforex |
//+------------------------------------------------------------------+

#property version   "1.00"
#property indicator_chart_window
#property indicator_buffers 2 + 11 + 1 + 1 + 1
#property indicator_plots 2

input int prd = 5; // Pivot Period
enum ENUM_PIVOT_SOURCE
  {
   PIVOT_SOURCE_CLOSE, // Close
   PIVOT_SOURCE_HIGH_LOW // High/Low
  };
input ENUM_PIVOT_SOURCE source = PIVOT_SOURCE_CLOSE; // Source for Pivot Points
enum ENUM_DIVERGENCE_TYPE
  {
   DIVERGENCE_TYPE_REGULAR, // Regular
   DIVERGENCE_TYPE_HIDDEN, // Hidden
   DIVERGENCE_TYPE_REGULAR_HIDDEN // Regular/Hidden
  };
input ENUM_DIVERGENCE_TYPE searchdiv = DIVERGENCE_TYPE_REGULAR; // Divergence Type
enum ENUM_INDICATOR_NAME_TYPE
  {
   INDICATOR_NAME_TYPE_FULL, // Full
   INDICATOR_NAME_TYPE_FIRST_LETTER, // First Letter
   INDICATOR_NAME_TYPE_DONT_SHOW // Don't Show
  };
input ENUM_INDICATOR_NAME_TYPE showindis = INDICATOR_NAME_TYPE_FULL; // Show Indicator Names
input int showlimit = 1; // Minimum Number of Divergence
input int maxpp = 10; // Maximum Pivot Points to Check
input int maxbars = 100; // Maximum Bars to Check
input bool shownum = true; // Show Divergence Number
input bool showlast = false; // Show Only Last Divergence
input bool dontconfirm = false; // Don't Wait for Confirmation
input bool showlines = true; // Show Divergence Lines
input bool showpivot = false; // Show Pivot Points
input bool calcmacd = true; // MACD
input bool calcmacda = true; // MACD Histogram
input bool calcrsi = true; // RSI
input bool calcstoc = true; // Stochastic
input bool calccci = true; // CCI
input bool calcmom = true; // Momentum
input bool calcobv = true; // OBV
input bool calcvwmacd = true; // VWmacd
input bool calccmf = true; // Chaikin Money Flow
input bool calcmfi = true; // Money Flow Index
input color pos_reg_div_col = clrYellow; // Positive Regular Divergence
input color neg_reg_div_col = clrAqua; // Negative Regular Divergence
input color pos_hid_div_col = clrLime; // Positive Hidden Divergence
input color neg_hid_div_col = clrRed; // Negative Hidden Divergence
input color pos_div_text_col = clrYellow; // Positive Divergence Text Color
input color neg_div_text_col = clrAqua; // Negative Divergence Text Color
input ENUM_LINE_STYLE reg_div_l_style_ = STYLE_SOLID; // Regular Divergence Line Style
input ENUM_LINE_STYLE hid_div_l_style_ = STYLE_DASH; // Hdden Divergence Line Style
input int reg_div_l_width = 2; // Regular Divergence Line Width
input int hid_div_l_width = 1; // Hidden Divergence Line Width
input bool showmas = false; // Show MAs 50 & 200
input color cma1col = clrLime; // MAs Color Bullish
input color cma2col = clrRed; // MAs Color Bearish

string inpUniqueId = "DV4";

int sma50Handle = INVALID_HANDLE, sma200Handle = INVALID_HANDLE;

#property indicator_label1 "SMA - 50"
#property indicator_type1 DRAW_LINE
#property indicator_style1 STYLE_SOLID
//#property indicator_color1 clrLime
#property indicator_width1 1

#property indicator_label2 "SMA - 200"
#property indicator_type2 DRAW_LINE
#property indicator_style2 STYLE_SOLID
//#property indicator_color2 clrRed
#property indicator_width2 1

double sma50[], sma200[];

double rsi[], macd[], signal[], deltamacd[], moment[], cci[], Obv[], stk[], vwmacd[], cmf[], Mfi[];
int rsiHandle = INVALID_HANDLE, macdHandle = INVALID_HANDLE, momentHandle = INVALID_HANDLE, cciHandle = INVALID_HANDLE,
    ObvHandle = INVALID_HANDLE, stkHandle = INVALID_HANDLE, MfiHandle = INVALID_HANDLE;

string indicators_name[11];
color div_colors[5];

#include <Arrays\ArrayInt.mqh>
#include <Arrays\ArrayDouble.mqh>
int maxarraysize = 20;
CArrayInt ph_positions, _ph_positions;
CArrayInt pl_positions, _pl_positions;
CArrayDouble ph_vals, _ph_vals;
CArrayDouble pl_vals, _pl_vals;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int array_get(const CArrayInt &array, const int x)
  {
   return(array.At(x));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double array_get(const CArrayDouble &array, const int x)
  {
   return(array.At(x));
  }

struct positive_regular_positive_hidden_divergence_struct
  {
   double            prsc[];
   int               SIZE;
   int               call(const double &src[], const double &close[], const double &low[], const int cond, const int i, const int rates_total)
     {
      int divlen = 0;
      if(SIZE <= rates_total)
        {
         SIZE = ArrayResize(prsc, rates_total + 500, 200);
        }
      prsc[i] = source == PIVOT_SOURCE_CLOSE ? close[i] : low[i];
      // if indicators higher than last value and close price is higher than las close
      if(i > 0)
         if(dontconfirm || src[i] > src[i - 1] || close[i] > close[i - 1])
           {
            int startpoint = dontconfirm ? 0 : 1; // don't check last candle
            // we search last 15 PPs
            const int bar_index = i;
            for(int x = 0 ; x <= maxpp - 1; x++)
              {
               int len = bar_index - array_get(pl_positions, x) + prd;
               // if we reach non valued array element or arrived 101. or previous bars then we don't search more
               if(array_get(pl_positions, x) == 0 || len > maxbars)
                  break;
               if(len > 5 &&
                  ((cond == 1 && src[i - startpoint] > src[i - len] && prsc[i - startpoint] < (array_get(pl_vals, x))) ||
                   (cond == 2 && src[i - startpoint] < src[i - len] && prsc[i - startpoint] > (array_get(pl_vals, x)))))
                 {
                  double slope1 = (src[i - startpoint] - src[i - len]) / fabs((i - len) - (i - startpoint)),
                         virtual_line1 = src[i - startpoint] - slope1,
                         slope2 = (close[i - startpoint] - close[i - len]) / fabs((i - len) - (i - startpoint)),
                         virtual_line2 = close[i - startpoint] - slope2;
                  bool arrived = true;
                  for(int y = 1 + startpoint ; y <= len - 1; y++)
                    {
                     if(src[i - y] < virtual_line1 || (close[i - y]) < virtual_line2)
                       {
                        arrived = false;
                        break;
                       }
                     virtual_line1 = virtual_line1 - slope1;
                     virtual_line2 = virtual_line2 - slope2;
                    }
                  if(arrived)
                    {
                     divlen = i - len;
                     break;
                    }
                 }
              }
           }
      return divlen;
     }
  } positive_regular_positive_hidden_divergence[22];

struct negative_regular_negative_hidden_divergence_struct
  {
   double            prsc[];
   int               SIZE;
   int               call(const double &src[], const double &close[], const double &high[], const int cond, const int i, const int rates_total)
     {
      int divlen = 0;
      if(SIZE <= rates_total)
        {
         SIZE = ArrayResize(prsc, rates_total + 500, 200);
        }
      prsc[i] = source == PIVOT_SOURCE_CLOSE ? close[i] : high[i];
      // if indicators higher than last value and close price is higher than las close
      if(i > 0)
         if(dontconfirm || src[i] < src[i - 1] || close[i] < close[i - 1])
           {
            int startpoint = dontconfirm ? 0 : 1; // don't check last candle
            // we search last 15 PPs
            const int bar_index = i;
            for(int x = 0 ; x <= maxpp - 1; x++)
              {
               //#define array_get(array, x) array.At(x)
               int len = bar_index - array_get(ph_positions, x) + prd;
               // if we reach non valued array element or arrived 101. or previous bars then we don't search more
               if(array_get(ph_positions, x) == 0 || len > maxbars)
                  break;
               if(len > 5 &&
                  ((cond == 1 && src[i - startpoint] < src[i - len] && prsc[i - startpoint] > (array_get(ph_vals, x))) ||
                   (cond == 2 && src[i - startpoint] > src[i - len] && prsc[i - startpoint] < (array_get(ph_vals, x)))))
                 {
                  double slope1 = (src[i - startpoint] - src[i - len]) / fabs((i - len) - (i - startpoint)),
                         virtual_line1 = src[i - startpoint] - slope1,
                         slope2 = (close[i - startpoint] - close[i - len]) / fabs((i - len) - (i - startpoint)),
                         virtual_line2 = close[i - startpoint] - slope2;
                  bool arrived = true;
                  for(int y = 1 + startpoint ; y <= len - 1; y++)
                    {
                     if(src[i - y] > virtual_line1 || (close[i - y]) > virtual_line2)
                       {
                        arrived = false;
                        break;
                       }
                     virtual_line1 = virtual_line1 - slope1;
                     virtual_line2 = virtual_line2 - slope2;
                    }
                  if(arrived)
                    {
                     divlen = i - len;
                     break;
                    }
                 }
              }
           }
      return divlen;
     }
  } negative_regular_negative_hidden_divergence[22];

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CArrayInt* calculate_divs(const bool cond, const double &indicator[], const double &close[], const double &low[], const double &high[], const int i, const int rates_total, int instance = 0)
  {
   static CArrayInt divs;
   divs.Clear();
   instance *= 2;
   divs.Add(cond && (searchdiv == DIVERGENCE_TYPE_REGULAR || searchdiv == DIVERGENCE_TYPE_REGULAR_HIDDEN) ? positive_regular_positive_hidden_divergence[instance + 0].call(indicator, close, low, 1, i, rates_total) : 0);
   divs.Add(cond && (searchdiv == DIVERGENCE_TYPE_REGULAR || searchdiv == DIVERGENCE_TYPE_REGULAR_HIDDEN) ? negative_regular_negative_hidden_divergence[instance + 0].call(indicator, close, high, 1, i, rates_total) : 0);
   divs.Add(cond && (searchdiv == DIVERGENCE_TYPE_HIDDEN || searchdiv == DIVERGENCE_TYPE_REGULAR_HIDDEN)  ? positive_regular_positive_hidden_divergence[instance + 1].call(indicator, close, low, 2, i, rates_total) : 0);
   divs.Add(cond && (searchdiv == DIVERGENCE_TYPE_HIDDEN || searchdiv == DIVERGENCE_TYPE_REGULAR_HIDDEN)  ? negative_regular_negative_hidden_divergence[instance + 1].call(indicator, close, high, 2, i, rates_total) : 0);
   return &divs;
  }

CArrayInt all_divergences, _all_divergences;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_set(CArrayInt &array, const int idx, const int value)
  {
   return(array.Update(idx, value));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_set(CArrayDouble &array, const int idx, const double value)
  {
   return(array.Update(idx, value));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void array_set_divs(CArrayInt *div_pointer, const int index)
  {
   if(all_divergences.Total() == 0)
      for(int x = 0; x < 44; x++)
        {
         all_divergences.Add(0);
        }
   for(int x = 0 ; x <= 3; x++)
      array_set(all_divergences, index * 4 + x, array_get(div_pointer, x));
  }

bool initFailed = false;
double strength[], type[], level[];

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_unshift(CArrayInt &array, const int value)
  {
   return(array.Insert(value, 0));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_unshift(CArrayDouble &array, const double value)
  {
   return(array.Insert(value, 0));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_pop(CArrayInt &array)
  {
   return(array.Delete(array.Total() - 1));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_pop(CArrayDouble &array)
  {
   return(array.Delete(array.Total() - 1));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int array_size(CArrayInt &array)
  {
   return(array.Total());
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int array_size(CArrayDouble &array)
  {
   return(array.Total());
  }

#include <Arrays\ArrayObj.mqh>
class Line : public CObject
  {
public:
                     Line(const string _name, const double _sprice, const datetime _sTime, const double _eprice, const datetime _eTime, const color _clr, const ENUM_LINE_STYLE _style, const int _width):
                     name(_name), sprice(_sprice), sTime(_sTime), eprice(_eprice), eTime(_eTime), clr(_clr), style(_style), width(_width) {}
                    ~Line(void) {}
   string            name;
   double            sprice, eprice;
   const datetime    sTime, eTime;
   color             clr;
   ENUM_LINE_STYLE   style;
   int               width;
  };
CArrayObj pos_div_lines, neg_div_lines;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void array_clear(CArrayObj &array)
  {
   array.Clear();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_old_pos_div_lines()
  {
   array_clear(pos_div_lines);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class Label : public CObject
  {
public:
                     Label(const string _name, const double _price, const datetime _time, const color _clr, const string &_text[], const ENUM_ANCHOR_POINT _anc, const double _range):
                     name(_name), price(_price), time(_time), clr(_clr),/* text(_text),*/ anc(_anc), range(_range) { ArrayCopy(text, _text); }
                    ~Label(void) {}
   string            name;
   double            price, range;
   datetime          time;
   color             clr;
   string            text[11];
   ENUM_ANCHOR_POINT anc;
  };
CArrayObj pos_div_labels, neg_div_labels;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_old_neg_div_lines()
  {
   array_clear(neg_div_lines);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_old_pos_div_labels()
  {
   array_clear(pos_div_labels);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_old_neg_div_labels()
  {
   array_clear(neg_div_labels);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_pop(CArrayObj &array)
  {
   return(array.Delete(array.Total() - 1));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int array_size(CArrayObj &array)
  {
   return(array.Total());
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_last_pos_div_lines_label(const int n)
  {
   if(n > 0 && array_size(pos_div_lines) >= n)
     {
      int asz = array_size(pos_div_lines);
      for(int j = 1; j <= n; j++)
         array_pop(pos_div_lines);
      if(array_size(pos_div_labels) > 0)
         array_pop(pos_div_labels);
     }
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void delete_last_neg_div_lines_label(const int n)
  {
   if(n > 0 && array_size(neg_div_lines) >= n)
     {
      int asz = array_size(neg_div_lines);
      for(int j = 1; j <= n; j++)
         array_pop(neg_div_lines);
      if(array_size(neg_div_labels) > 0)
         array_pop(neg_div_labels);
     }
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_includes(CArrayInt &_array, const int element)
  {
   CArrayInt array;
   array.Clear();
   array.AssignArray(&_array);
   array.Sort();
   int search = array.Search(element);

   return(search != -1);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_push(CArrayObj &array, CObject *element)
  {
   return(array.Add(element));
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool array_push(CArrayInt &array, const int element)
  {
   return(array.Add(element));
  }
//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
int OnInit()
  {
   int idx = 0;
//--- indicator buffers mapping
   SetIndexBuffer(idx++, sma50, INDICATOR_DATA);
   SetIndexBuffer(idx++, sma200, INDICATOR_DATA);
   SetIndexBuffer(idx++, rsi, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, macd, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, signal, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, deltamacd, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, moment, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, cci, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, Obv, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, stk, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, vwmacd, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, cmf, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, Mfi, INDICATOR_CALCULATIONS);
   SetIndexBuffer(idx++, strength, INDICATOR_CALCULATIONS); // 13
   SetIndexBuffer(idx++, type, INDICATOR_CALCULATIONS); // 14
   SetIndexBuffer(idx++, level, INDICATOR_CALCULATIONS); // 15
   PlotIndexSetInteger(0, PLOT_LINE_COLOR, showmas ? cma1col : clrNONE);
   PlotIndexSetInteger(1, PLOT_LINE_COLOR, showmas ? cma2col : clrNONE);
//---
   sma50Handle = iMA(_Symbol, PERIOD_CURRENT, 50, 0, MODE_SMA, PRICE_CLOSE);
   sma200Handle = iMA(_Symbol, PERIOD_CURRENT, 200, 0, MODE_SMA, PRICE_CLOSE);
   rsiHandle = iRSI(_Symbol, PERIOD_CURRENT, 14, PRICE_CLOSE);
   macdHandle = iMACD(_Symbol, PERIOD_CURRENT, 12, 26, 9, PRICE_CLOSE);
   momentHandle = iMomentum(_Symbol, PERIOD_CURRENT, 10, PRICE_CLOSE);
   cciHandle = iCCI(_Symbol, PERIOD_CURRENT, 10, PRICE_CLOSE);
   ObvHandle = iOBV(_Symbol, PERIOD_CURRENT, VOLUME_TICK);
   stkHandle = iStochastic(_Symbol, PERIOD_CURRENT, 14, 3, 3, MODE_SMA, STO_LOWHIGH);
   MfiHandle = iMFI(_Symbol, PERIOD_CURRENT, 14, VOLUME_TICK);
   if(sma50Handle == INVALID_HANDLE || sma200Handle == INVALID_HANDLE ||
      rsiHandle == INVALID_HANDLE || macdHandle == INVALID_HANDLE || momentHandle == INVALID_HANDLE ||
      cciHandle == INVALID_HANDLE || ObvHandle == INVALID_HANDLE || stkHandle == INVALID_HANDLE || MfiHandle == INVALID_HANDLE)
     {
      Alert("Failed to create handles, Try again");
      return(INIT_FAILED);
     }
   //indicators_name.Reserve(11);
   //div_colors.Reserve(4);
// names
   indicators_name[0] = (showindis == INDICATOR_NAME_TYPE_FULL ? "MACD" : "M");
   indicators_name[1] = (showindis == INDICATOR_NAME_TYPE_FULL ? "Hist" : "H");
   indicators_name[2] = (showindis == INDICATOR_NAME_TYPE_FULL ? "RSI" : "E");
   indicators_name[3] = (showindis == INDICATOR_NAME_TYPE_FULL ? "Stoch" : "S");
   indicators_name[4] = (showindis == INDICATOR_NAME_TYPE_FULL ? "CCI" : "C");
   indicators_name[5] = (showindis == INDICATOR_NAME_TYPE_FULL ? "MOM" : "M");
   indicators_name[6] = (showindis == INDICATOR_NAME_TYPE_FULL ? "OBV" : "O");
   indicators_name[7] = (showindis == INDICATOR_NAME_TYPE_FULL ? "VWMACD" : "V");
   indicators_name[8] = (showindis == INDICATOR_NAME_TYPE_FULL ? "CMF" : "C");
   indicators_name[9] = (showindis == INDICATOR_NAME_TYPE_FULL ? "MFI" : "M");
   indicators_name[10] = (showindis == INDICATOR_NAME_TYPE_FULL ? "Extrn" : "X");
//colors
   //div_colors.Push(pos_reg_div_col);
   //div_colors.Push(neg_reg_div_col);
   //div_colors.Push(pos_hid_div_col);
   //div_colors.Push(neg_hid_div_col);
   
   div_colors[0] = pos_reg_div_col;
   div_colors[1] = neg_reg_div_col;
   div_colors[2] = pos_hid_div_col;
   div_colors[3] = neg_hid_div_col;
   return(INIT_SUCCEEDED);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   ObjectsDeleteAll(0, inpUniqueId + ":");
  }
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,
                const int prev_calculated,
                const datetime &time[],
                const double &open[],
                const double &high[],
                const double &low[],
                const double &close[],
                const long &tick_volume[],
                const long &volume[],
                const int &spread[])
  {
   if(initFailed)
     {
      return prev_calculated;
     }
   if(prev_calculated <= 0)
     {
      ph_positions.Clear();
      pl_positions.Clear();
      ph_vals.Clear();
      pl_vals.Clear();
      ObjectsDeleteAll(0, inpUniqueId + ":");

      all_divergences.Clear();

      _ph_positions.Clear();
      _pl_positions.Clear();
      _ph_vals.Clear();
      _pl_vals.Clear();
      _all_divergences.Clear();
     }
   if(true)
     {
      ph_positions.AssignArray(&_ph_positions);
      pl_positions.AssignArray(&_pl_positions);
      ph_vals.AssignArray(&_ph_vals);
      pl_vals.AssignArray(&_pl_vals);
      all_divergences.AssignArray(&_all_divergences);
     }
   int limit = (prev_calculated > 0) ? prev_calculated - 1 : 0;
   int toCopy = (prev_calculated > 0) ? rates_total - prev_calculated + 1 : rates_total;
   if(CopyBuffer(sma50Handle, MAIN_LINE, 0, toCopy, sma50) != toCopy ||
      CopyBuffer(sma200Handle, MAIN_LINE, 0, toCopy, sma200) != toCopy ||
      CopyBuffer(rsiHandle, MAIN_LINE, 0, toCopy, rsi) != toCopy ||
      CopyBuffer(macdHandle, MAIN_LINE, 0, toCopy, macd) != toCopy ||
      CopyBuffer(macdHandle, SIGNAL_LINE, 0, toCopy, signal) != toCopy ||
      CopyBuffer(momentHandle, MAIN_LINE, 0, toCopy, moment) != toCopy ||
      CopyBuffer(cciHandle, MAIN_LINE, 0, toCopy, cci) != toCopy ||
      CopyBuffer(ObvHandle, MAIN_LINE, 0, toCopy, Obv) != toCopy ||
      CopyBuffer(stkHandle, MAIN_LINE, 0, toCopy, stk) != toCopy ||
      CopyBuffer(MfiHandle, MAIN_LINE, 0, toCopy, Mfi) != toCopy)
     {
      Print("CopyBuffer() failed, Trying again on next tick");
      return(prev_calculated);
     }
     
     
//---
   for(int i = limit; i < /*rates_total*/ rates_total /*- (dontconfirm ? 0 : 1)*/ && !_StopFlag; i++)
     {
      deltamacd[i] = macd[i] - signal[i];
      double maFast = iSma(close[i] * tick_volume[i], 12, i, rates_total, 0) / iSma(tick_volume[i], 12, i, rates_total, 1),
             maSlow = iSma(close[i] * tick_volume[i], 26, i, rates_total, 2) / iSma(tick_volume[i], 26, i, rates_total, 3);
      vwmacd[i] = maFast - maSlow;
      vwmacd[i] = 0;
      double Cmfm = (high[i] - low[i]) != 0 ? ((close[i] - low[i]) - (high[i] - close[i])) / (high[i] - low[i]) : 0,
             Cmfv = Cmfm * tick_volume[i];
      cmf[i] = iSma(Cmfv, 21, i, rates_total, 4) / iSma(tick_volume[i], 21, i, rates_total, 5);
      cmf[i] = 0;
      static int prdTotal = prd + prd + 1;
      int checkWith = i - prd,
          phIdx = source == PIVOT_SOURCE_CLOSE ? ArrayMaximum(close, MathMax(i - prdTotal + 1, 0), prdTotal) : ArrayMaximum(high, MathMax(i - prdTotal + 1, 0), prdTotal),
          plIdx = source == PIVOT_SOURCE_CLOSE ? ArrayMinimum(close, MathMax(i - prdTotal + 1, 0), prdTotal) : ArrayMinimum(low, MathMax(i - prdTotal + 1, 0), prdTotal);
      double ph = phIdx == checkWith ? source == PIVOT_SOURCE_CLOSE ? close[phIdx] : high[phIdx] : EMPTY_VALUE,
             pl = plIdx == checkWith ? source == PIVOT_SOURCE_CLOSE ? close[plIdx] : low[plIdx] : EMPTY_VALUE;
      if(ph != EMPTY_VALUE && showpivot)
         CreateText(/*inpUniqueId + ":" + */"ph" + ":" + TimeToString(time[phIdx]), ph, time[phIdx], clrRed, "H", ANCHOR_LOWER);
      if(pl != EMPTY_VALUE && showpivot)
         CreateText(/*inpUniqueId + ":" + */"pl" + ":" + TimeToString(time[plIdx]), pl, time[plIdx], clrLime, "L", ANCHOR_UPPER);
      const int bar_index = i;
      if(ph != EMPTY_VALUE)
        {
         array_unshift(ph_positions, bar_index);
         array_unshift(ph_vals, ph);
         if(array_size(ph_positions) > maxarraysize)
           {
            array_pop(ph_positions);
            array_pop(ph_vals);
           }
        }
      // add PLs to the array
      if(pl != EMPTY_VALUE)
        {
         array_unshift(pl_positions, bar_index);
         array_unshift(pl_vals, pl);
         if(array_size(pl_positions) > maxarraysize)
           {
            array_pop(pl_positions);
            array_pop(pl_vals);
           }
        }
      array_set_divs(calculate_divs(calcmacd, macd, close, low, high, i, rates_total, 0), 0);
      array_set_divs(calculate_divs(calcmacda, deltamacd, close, low, high, i, rates_total, 1), 1);
      array_set_divs(calculate_divs(calcrsi, rsi, close, low, high, i, rates_total, 2), 2);
      array_set_divs(calculate_divs(calcstoc, stk, close, low, high, i, rates_total, 3), 3);
      array_set_divs(calculate_divs(calccci, cci, close, low, high, i, rates_total, 4), 4);
      array_set_divs(calculate_divs(calcmom, moment, close, low, high, i, rates_total, 5), 5);
      array_set_divs(calculate_divs(calcobv, Obv, close, low, high, i, rates_total, 6), 6);
      array_set_divs(calculate_divs(calcvwmacd, vwmacd, close, low, high, i, rates_total, 7), 7);
      array_set_divs(calculate_divs(calccmf, cmf, close, low, high, i, rates_total, 8), 8);
      array_set_divs(calculate_divs(calcmfi, Mfi, close, low, high, i, rates_total, 9), 9);
      //array_set_divs(calculate_divs(calcext, externalindi), 10)
#define sign(val) val > 0 ? 1 : val < 0 ? -1 : 0
      int total_div = 0;
      for(int x = 0 ; x <= array_size(all_divergences) - 1; x++)
         total_div = total_div + (int)round(sign(array_get(all_divergences, x)));
      if(total_div < showlimit)
         for(int x = 0; x < array_size(all_divergences); x++)
            all_divergences.Update(x, 0);

      static int last_pos_div_lines = 0, _last_pos_div_lines = 0;
      static int last_neg_div_lines = 0, _last_neg_div_lines = 0;
      static bool remove_last_pos_divs = false, _remove_last_pos_divs = false;
      static bool remove_last_neg_divs = false, _remove_last_neg_divs = false;
      if(i == 0)
        {
         last_pos_div_lines = _last_pos_div_lines = 0;
         last_neg_div_lines = _last_neg_div_lines = 0;
         remove_last_pos_divs = _remove_last_pos_divs = false;
         remove_last_neg_divs = _remove_last_neg_divs = false;
        }
      last_pos_div_lines = _last_pos_div_lines;
      last_neg_div_lines = _last_neg_div_lines;
      remove_last_pos_divs = _remove_last_pos_divs;
      remove_last_neg_divs = _remove_last_neg_divs;

      if(pl != EMPTY_VALUE)
        {
         remove_last_pos_divs = false;
         last_pos_div_lines = 0;
        }
      if(ph != EMPTY_VALUE)
        {
         remove_last_neg_divs = false;
         last_neg_div_lines = 0;
        }

      string divergence_text_top[11] { NULL }, divergence_text_bottom[11] { NULL };
      static const string divType[]{ "reg_pos", "reg_neg", "hid_pos", "hid_neg" };
      string divType_ = "";
      double level_ = 0;
      static CArrayInt distances;
      distances.Clear();
      int dnumdiv_top = 0;
      int dnumdiv_bottom = 0;
      color top_label_col = clrWhite;
      color bottom_label_col = clrWhite;
      bool old_pos_divs_can_be_removed = true;
      bool old_neg_divs_can_be_removed = true;
      static const int startpoint = dontconfirm ? 0 : 1;
      for(int x = 0; x < 10; x++)
        {
         int div_type = -1;
         for(int y = 0; y <= 3; y++)
           {
            if(array_get(all_divergences, x * 4 + y) > 0)
              {
               div_type = y;
               if((y % 2) == 1)
                 {
                  dnumdiv_top = dnumdiv_top + 1;
                  top_label_col = div_colors[y];
                 }
               if((y % 2) == 0)
                 {
                  dnumdiv_bottom = dnumdiv_bottom + 1;
                  bottom_label_col = div_colors[y];
                 }
               if(!array_includes(distances, array_get(all_divergences, x * 4 + y)))
                 {
                  array_push(distances, array_get(all_divergences, x * 4 + y));
//---
                  divType_ = divType[y];
                  level_ = (source == PIVOT_SOURCE_CLOSE ?
                           close[bar_index - startpoint] :
                           (y % 2) == 0 ?
                           low[bar_index - startpoint] :
                           high[bar_index - startpoint]);
//---
                  Line *new_line = showlines ?
                                   new Line(
                                      inpUniqueId + ":" + IntegerToString(array_get(all_divergences, x * 4 + y)) + ":" + TimeToString(time[bar_index - startpoint]),

                                      (source == PIVOT_SOURCE_CLOSE ?
                                       close[array_get(all_divergences, x * 4 + y)] :
                                       (y % 2) == 0 ?
                                       low[array_get(all_divergences, x * 4 + y)] :
                                       high[array_get(all_divergences, x * 4 + y)]),

                                      time[array_get(all_divergences, x * 4 + y)],

                                      (source == PIVOT_SOURCE_CLOSE ?
                                       close[bar_index - startpoint] :
                                       (y % 2) == 0 ?
                                       low[bar_index - startpoint] :
                                       high[bar_index - startpoint]),

                                      time[bar_index - startpoint],

                                      div_colors[y],

                                      y < 2 ? reg_div_l_style_ : hid_div_l_style_,

                                      y < 2 ? reg_div_l_width : hid_div_l_width
                                   )
                                   : NULL;
                  if((y % 2) == 0)
                    {
                     if(old_pos_divs_can_be_removed)
                       {
                        old_pos_divs_can_be_removed = false;
                        if(! showlast && remove_last_pos_divs)
                          {
                           delete_last_pos_div_lines_label(last_pos_div_lines);
                           last_pos_div_lines = 0;
                          }
                        if(showlast)
                           delete_old_pos_div_lines();
                       }
                     array_push(pos_div_lines, new_line);
                     last_pos_div_lines = last_pos_div_lines + 1;
                     remove_last_pos_divs = true;
                    }
                  if((y % 2) == 1)
                    {
                     if(old_neg_divs_can_be_removed)
                       {
                        old_neg_divs_can_be_removed = false;
                        if(! showlast && remove_last_neg_divs)
                          {
                           delete_last_neg_div_lines_label(last_neg_div_lines);
                           last_neg_div_lines = 0;
                          }
                        if(showlast)
                           delete_old_neg_div_lines();
                       }
                     array_push(neg_div_lines, new_line);
                     last_neg_div_lines = last_neg_div_lines + 1;
                     remove_last_neg_divs = true;
                    }
                 }
              }
           }

         if(div_type >= 0)
           {
            divergence_text_top[x + 1] = ((div_type % 2) == 1 ? (showindis != INDICATOR_NAME_TYPE_DONT_SHOW ? indicators_name[x] : "") : "");
            divergence_text_bottom[x + 1] = ((div_type % 2) == 0 ? (showindis != INDICATOR_NAME_TYPE_DONT_SHOW ? indicators_name[x] : "") : "");
           }
        }
      divergence_text_top[0] = shownum ? dnumdiv_top > 0 ? IntegerToString(dnumdiv_top) : "" : "";
      divergence_text_bottom[0] = shownum ? dnumdiv_bottom > 0 ? IntegerToString(dnumdiv_bottom) : "" : "";
      bool divergence_text_top_valid = false, divergence_text_bottom_valid = false;
      for(int j = 0; j < 11; j++)
        {
         if(divergence_text_top[j] != "" && divergence_text_top[j] != NULL)
           {
            divergence_text_top_valid = true;
           }
         if(divergence_text_bottom[j] != "" && divergence_text_bottom[j] != NULL)
           {
            divergence_text_bottom_valid = true;
           }
        }

      double range = high[i] - low[i], mul = 0;
      for(int x = 0; x < 50 && (i - x) >= 0; x++)
        {
         range += (high[i - x] - low[i - x]);
        }
      range /= 50;

      if(divergence_text_top_valid)
        {
         if(showlast)
            delete_old_neg_div_labels();
         array_push(neg_div_labels,
                    new Label(
                       inpUniqueId + ":" + "top" + ":" + TimeToString(time[bar_index - startpoint]),
                       MathMax(high[bar_index], high[bar_index - 1]),
                       time[bar_index],
                       neg_div_text_col,
                       divergence_text_top,
                       ANCHOR_LOWER,
                       range));
        }

      if(divergence_text_bottom_valid)
        {
         if(showlast)
            delete_old_pos_div_labels();
         array_push(pos_div_labels,
                    new Label(
                       inpUniqueId + ":" + "bot" + ":" + TimeToString(time[bar_index - startpoint]),
                       MathMin(low[bar_index], low[bar_index - 1]),
                       time[bar_index],
                       pos_div_text_col,
                       divergence_text_bottom,
                       ANCHOR_UPPER,
                       range));
        }

      if(i < rates_total - 1)
        {
         _ph_positions.AssignArray(&ph_positions);
         _pl_positions.AssignArray(&pl_positions);
         _ph_vals.AssignArray(&ph_vals);
         _pl_vals.AssignArray(&pl_vals);
         _all_divergences.AssignArray(&all_divergences);

         _last_pos_div_lines = last_pos_div_lines;
         _last_neg_div_lines = last_neg_div_lines;
         _remove_last_pos_divs = remove_last_pos_divs;
         _remove_last_neg_divs = remove_last_neg_divs;
        }
        
      strength[i] = MathAbs(dnumdiv_top - dnumdiv_bottom);
      
      if(dnumdiv_bottom > 0)
      {
         strength[i] = -strength[i];
      }
      
      type[i] = divType_ == divType[0] ? 0 : divType_ == divType[1] ? 1 : divType_ == divType[2] ? 2 : divType_ == divType[3] ? 3 : -1;
      level[i] = level_;
     }

   if(rates_total != prev_calculated)
     {
      //if(type[rates_total - 2] != -1)
      //   PrintFormat("strength = %f | type = %f | level = %f", strength[rates_total - 2], type[rates_total - 2], level[rates_total - 2]);
      ObjectsDeleteAll(0, inpUniqueId + ":");
      for(int i = 0; i < pos_div_lines.Total(); i++)
        {
         Line *line = pos_div_lines.At(i);
         CreateLines(line.name, 0, line.sprice, line.sTime, line.eprice, line.eTime, line.clr, line.style, line.width);
        }
      for(int i = 0; i < neg_div_lines.Total(); i++)
        {
         Line *line = neg_div_lines.At(i);
         CreateLines(line.name, 0, line.sprice, line.sTime, line.eprice, line.eTime, line.clr, line.style, line.width);
        }
      for(int i = 0; i < pos_div_labels.Total(); i++)
        {
         Label *label = pos_div_labels.At(i);
         int mult = 0;
         for(int j = 0; j < 11; j++)
           {
            if(label.text[j] != "" && label.text[j] != NULL)
              {
               CreateText(label.name + ":" + IntegerToString(j), label.price - label.range * mult++, label.time, label.clr, label.text[j], label.anc);
              }
           }
        }
      for(int i = 0; i < neg_div_labels.Total(); i++)
        {
         Label *label = neg_div_labels.At(i);
         int mult = 0;
         for(int j = 0; j < 11; j++)
           {
            if(label.text[j] != "" && label.text[j] != NULL)
              {
               CreateText(label.name + ":" + IntegerToString(j), label.price + label.range * mult++, label.time, label.clr, label.text[j], label.anc);
              }
           }
        }
      //ChartRedraw();
     }
//--- return value of prev_calculated for next call
   return(rates_total);
  }
//+------------------------------------------------------------------+
double workSma[][6];
double iSma(double price, int period, int r, int _bars, int instanceNo = 0)
  {
   static int SIZE = -1;
   if(SIZE <= _bars)
     {
      ArrayResize(workSma, _bars + 500, 2000);
      SIZE = ArrayRange(workSma, 0);
     }
   workSma[r][instanceNo + 0] = price;
   double avg = price;
   int k = 1;
   for(; k < period && (r - k) >= 0; k++)
      avg += workSma[r - k][instanceNo + 0];
   return(avg / (double)k);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CreateText(const string name, const double price, const datetime time, const color clr, const string text, const ENUM_ANCHOR_POINT anc)
  {
   bool create = ObjectCreate(0, name, OBJ_TEXT, 0, 0, 0);
   ObjectSetDouble(0, name, OBJPROP_PRICE, price);
   ObjectSetInteger(0, name, OBJPROP_TIME, time);
   ObjectSetInteger(0, name, OBJPROP_COLOR, clr);
   ObjectSetString(0, name, OBJPROP_TEXT, text);
   ObjectSetString(0, name, OBJPROP_FONT, "Tahoe");
   ObjectSetInteger(0, name, OBJPROP_FONTSIZE, 8);
   ObjectSetInteger(0, name, OBJPROP_ANCHOR, anc);
   ObjectSetInteger(0, name, OBJPROP_SELECTABLE, false);
   ObjectSetInteger(0, name, OBJPROP_SELECTED, false);
   return(create);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CreateLines(const string name, const int wnd, const double sprice, const datetime sTime, const double eprice, const datetime eTime, const color clr, const ENUM_LINE_STYLE style, const int width, const string tooltip = "", const bool rayRight = false)
  {
   bool create = ObjectCreate(0, name, OBJ_TREND, wnd, 0, 0, 0);
   ObjectSetDouble(0, name, OBJPROP_PRICE, 0, sprice);
   ObjectSetDouble(0, name, OBJPROP_PRICE, 1, eprice);
   ObjectSetInteger(0, name, OBJPROP_TIME, 0, sTime);
   ObjectSetInteger(0, name, OBJPROP_TIME, 1, eTime);
   ObjectSetInteger(0, name, OBJPROP_RAY_RIGHT, rayRight);
   ObjectSetInteger(0, name, OBJPROP_COLOR, clr);
   ObjectSetInteger(0, name, OBJPROP_STYLE, style);
   ObjectSetInteger(0, name, OBJPROP_WIDTH, width);
   ObjectSetInteger(0, name, OBJPROP_SELECTABLE, false);
   ObjectSetInteger(0, name, OBJPROP_SELECTED, false);
   if(tooltip != "")
      ObjectSetString(0, name, OBJPROP_TOOLTIP, tooltip);
   return(create);
  }
//+------------------------------------------------------------------+
